using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using static System.Diagnostics.Stopwatch;

namespace Stopper {
	public class MorphFinder : MVRScript {
		
		MeshCollider meshCollider;
		GameObject meshColliderContainer;
		
		int previewVertex = 0;
		int selectedVertex = 0;
		
		int frame = 0;
		
		Vector3 selectedPoint;
		float selectedDistance; 
		
		List<MorphDelta> applicableMorphs;
		DAZCharacterRun characterRun;
		
		List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();
		Dictionary<int, int> morphVertices;
		
		List<JSONStorableFloat> jsonFloats = new List<JSONStorableFloat>();
		
		GameObject previewSphere;
		GameObject selectionSphere;
		LineDrawer lineDrawer;
		
		JSONStorableFloat maxMorphCount;
		JSONStorableBool showSelection;
		
		public class MorphDelta {
			public DAZMorph morph;
			public Vector3 delta;
			public float totalMagnitude = 0f;
			public float alignment = 0f;
		}
		
		public override void Init() {
			meshColliderContainer = CreateChildObject(containingAtom.gameObject, "MorphFinder");
			meshColliderContainer.layer = 1;
			createdObjects.Add(meshColliderContainer);
			
			characterRun = containingAtom.GetComponentInChildren<DAZCharacterRun>();
			
			DAZMergedSkinV2 mergedSkin = containingAtom.GetComponentInChildren<DAZMergedSkinV2>();
			ComputeBuffer buffer = mergedSkin.rawVertsBuffer;
			
			
			Vector3[] vertices = new Vector3[buffer.count];
			buffer.GetData(vertices,0,0,buffer.count);
			
			Material lineMaterial = new Material(Shader.Find("Battlehub/RTGizmos/Handles"));
			lineMaterial.SetFloat("_Offset", 1f);
			lineMaterial.SetFloat("_MinAlpha", 1f);
			lineMaterial.color = Color.green;
			
			lineDrawer = new LineDrawer(lineMaterial);
			
			Mesh mesh = new Mesh();
			mesh.vertices = vertices;
			mesh.uv = mergedSkin.dazMesh.UV;
		    mesh.triangles = mergedSkin.dazMesh.baseTriangles;
			
			meshCollider = meshColliderContainer.AddComponent<MeshCollider>();
			meshCollider.sharedMesh = mesh;
			meshCollider.convex = false;
			meshCollider.enabled = true;
			
			selectionSphere = CreateSphere(Color.green);
			selectionSphere.GetComponent<MeshRenderer>().enabled = false;
			
			previewSphere = CreateSphere(Color.white);
			previewSphere.GetComponent<MeshRenderer>().enabled = false;
			
			showSelection = new JSONStorableBool("Show Selection", false);
			var showSelectionToggle = CreateToggle(showSelection, false);
			showSelectionToggle.toggle.onValueChanged.AddListener((val) => {
				if(val){
					RefreshMeshCollider();
				}
			});
			
			maxMorphCount = AddSlider("Max Morph Count", 20, 1, 50, null, true,true);
			
			UIDynamicButton refreshCollider = CreateButton("Refresh Mesh Collider");
			refreshCollider.button.onClick.AddListener(() => {
				RefreshMeshCollider();
			});
			
			CreateTextField(new JSONStorableString("PositiveMorphs", "Positive Morphs"), false).height = 10;
			CreateTextField(new JSONStorableString("NegativeMorphs", "Negative Morphs"), true).height = 10;
		}
		
		private void Update(){
			if(meshCollider == null){
				return;
			}
			
			Mesh mesh = meshCollider.sharedMesh;
			if(showSelection.val || Input.GetKey(KeyCode.LeftControl)){
				
				RaycastHit hit;
				Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
				if(meshCollider.Raycast(ray, out hit, 100f)){
					var vertex1 = mesh.triangles[hit.triangleIndex * 3 + 0];
					var vertex2 = mesh.triangles[hit.triangleIndex * 3 + 1];
					var vertex3 = mesh.triangles[hit.triangleIndex * 3 + 2];
					
					var vector1 = mesh.vertices[mesh.triangles[vertex1]];
					var vector2 = mesh.vertices[mesh.triangles[vertex2]];
					var vector3 = mesh.vertices[mesh.triangles[vertex3]];
					
					float min_distance = (hit.point - vector1).sqrMagnitude;
					previewVertex = vertex1;
					
					float distance = (hit.point - vector2).sqrMagnitude;
					if(distance < min_distance){
						previewVertex = vertex2;
					}
					
					distance = (hit.point - vector3).sqrMagnitude;
					if(distance < min_distance){
						previewVertex = vertex3;
					}
					
					selectedDistance = hit.distance;
					previewSphere.transform.position = mesh.vertices[previewVertex];
					previewSphere.GetComponent<MeshRenderer>().enabled = true;
				}
			}
			
			if (Input.GetKey(KeyCode.LeftControl)){
				
				if(selectedVertex == 0){
					selectedVertex = previewVertex;
					applicableMorphs = new List<MorphDelta>();
					
						
						
					selectionSphere.transform.position = mesh.vertices[selectedVertex];
					
					
					
					Vector3 startMousePos = Input.mousePosition;
					startMousePos.z = selectedDistance;
					Vector3 mouseWorldPosStart = Camera.main.ScreenToWorldPoint(startMousePos);
					selectedPoint = mesh.vertices[selectedVertex];
					
					
					ProcessMorphBank(characterRun.morphBank1);
					//ProcessMorphBank(characterRun.morphBank2);
					//ProcessMorphBank(characterRun.morphBank3);
					
					if(characterRun.useOtherGenderMorphs){
						ProcessMorphBank(characterRun.morphBank1OtherGender);
					}
				}
					
				Vector3 mousePos = Input.mousePosition;
				mousePos.z = selectedDistance;
				Vector3 mouseWorldPos = Camera.main.ScreenToWorldPoint(mousePos);
				Vector3 selectDir = Vector3.zero;
				if(mouseWorldPos != selectedPoint){
					
					selectDir = (mouseWorldPos - selectedPoint);
					if(selectDir.magnitude < 0.01f){
						selectDir = Vector3.zero;
						foreach(var md in applicableMorphs){
							md.alignment = md.delta.magnitude;
						}
					} else {
						selectDir = selectDir.normalized;
						foreach(var md in applicableMorphs){
							md.alignment = 0f;
						}
					}
					
					applicableMorphs.Sort( (m1,m2)=> {
						if(m1.alignment == 0f) m1.alignment = Vector3.Dot(m1.delta, selectDir);
						if(m2.alignment == 0f) m2.alignment = Vector3.Dot(m2.delta, selectDir);
						
						float relevance1 = Math.Abs(m1.alignment*100f) / m1.totalMagnitude;
						float relevance2 = Math.Abs(m2.alignment*100f) / m2.totalMagnitude;
						return relevance2.CompareTo(relevance1);
					});
				} 
				
				if(selectDir != Vector3.zero){
					lineDrawer.SetLinePoints(selectedPoint, mouseWorldPos);
					lineDrawer.Draw();
				}
				
				
				selectionSphere.GetComponent<MeshRenderer>().enabled = true;	
				previewSphere.GetComponent<MeshRenderer>().enabled = false;	
				
			} else if(selectedVertex != 0){
					
					foreach(var entry in sliderToJSONStorableFloat){
						if(entry.Value != maxMorphCount){
							RemoveSlider(entry.Key);
						}
					}
					jsonFloats.Clear();
					
					int numSliders = 0;
					
					foreach(var md in applicableMorphs){
						if(md.morph.resolvedDisplayName.StartsWith("MCM"))
							continue;
						
						var jsonFloat = AddSlider($"{md.morph.resolvedDisplayName}", md.morph.jsonFloat.defaultVal, -1f, 1f, (float val) => {
							md.morph.morphValue = val;
						}, true, md.alignment < 0);
						
						jsonFloat.val = md.morph.morphValue;
						
						jsonFloats.Add(jsonFloat);
						
						if(numSliders++ > maxMorphCount.val){
							break;
						}
					}
					
				selectedVertex = 0;
				selectionSphere.GetComponent<MeshRenderer>().enabled = false;		
			}
		}
		
		private void RefreshMeshCollider(){
			Mesh mesh = meshCollider.sharedMesh;
					
			DAZMergedSkinV2 mergedSkin = containingAtom.GetComponentInChildren<DAZMergedSkinV2>();
			ComputeBuffer buffer = mergedSkin.rawVertsBuffer;
		
			Vector3[] vertices = new Vector3[buffer.count];
			buffer.GetData(vertices,0,0,buffer.count);
			
			mesh.vertices = vertices;
			
			meshCollider.sharedMesh = null;
			meshCollider.sharedMesh = mesh;
		}
		
		private void ProcessMorphBank(DAZMorphBank morphBank){
			foreach(DAZMorph m in morphBank.morphs){
				if(m.deltas == null){
					continue;
				}
				MorphDelta md = new MorphDelta(){morph = m};
				md.totalMagnitude = 0f;
				
				foreach( DAZMorphVertex morphVert in m.deltas) {
					float deltaMagnitude = morphVert.delta.magnitude;
					md.totalMagnitude += deltaMagnitude;
					
					if(morphVert != null && m.morphName != null){
						if( (morphVert.vertex == selectedVertex)
							&& deltaMagnitude > 0.001f){
							md.delta = morphVert.delta;
							applicableMorphs.Add(md);
						}	
					}
				}
			}
		}
			
		
		private void OnDisable()
		{
			foreach(UnityEngine.Object gameObject in createdObjects) {
				UnityEngine.Object.Destroy(gameObject);
			}
		}
		
		private GameObject CreateChildObject(GameObject parent, string name) {
			var gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
			
			return gameObject;
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = rangeAdjust;
			
			if(onChange != null){
				storableFloat.setCallbackFunction = onChange;
			}
			
			return storableFloat;
		}
		
		private GameObject CreateSphere(Color color){
			var sphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
			//sphere.transform.parent = mainContainer.transform;
			//sphere.layer = 30;
			
			
			SphereCollider collider = sphere.GetComponent<SphereCollider>();
			UnityEngine.Object.DestroyImmediate(collider);
			
			MeshFilter meshFilter = sphere.GetComponent<MeshFilter>();
			Mesh mesh = meshFilter.mesh;
			var vertices = mesh.vertices;
			for(int i = 0; i < mesh.vertices.Length; i++){
				vertices[i] = vertices[i] * 0.005f;
			}
			mesh.vertices = vertices;
			
			sphere.transform.localPosition = new Vector3(0f,0f,0f);
			
			
			sphere.GetComponent<Renderer>().material.color = color;
			
			
			createdObjects.Add(sphere);
			
			return sphere;
		}
	}
}